// http://standards-oui.ieee.org/oui/oui.txt
oneoui(0x000c03, HDMI,         "HDMI"        )
oneoui(0xc45dd8, HDMIForum,    "HDMI Forum"  )
oneoui(0x90848b, HDR10,        "HDR10+"      )
oneoui(0x00001a, AMD,          "AMD"         )
oneoui(0x00044b, NVIDIA,       "NVIDIA"      )
oneoui(0x000c6e, ASUS,         "ASUS"        )
oneoui(0x0010fa, Apple,        "Apple"       )
oneoui(0x0014b9, MSTAR,        "MSTAR"       )
oneoui(0x00d046, Dolby,        "Dolby"       )
oneoui(0x00e047, InFocus,      "InFocus"     )
oneoui(0xca125c, Microsoft,    "Microsoft"   )

// http://standards-oui.ieee.org/cid/cid.txt
oneoui(0x3a0292, VESA,         "VESA"        )

// https://uefi.org/pnp_id_list
oneoui(0x41505000, asciiApple,   "Apple"     )  // 'APP\0'

#undef oneoui
